import warnings

try:
    import cv2
except ImportError:
    cv2 = None

import robothub

__all__ = ['HubPacket', 'DetectionPacket', 'TrackerPacket', 'DepthPacket', 'IMUPacket']


class HubPacket:
    """
    Base class for all packets. Represents a packet containing a frame.
    """

    def __init__(self, device: 'Device', packet):
        self.device = device
        self._packet = packet

        self.frame = self._packet.frame

    def upload_as_event(self, title: str):
        try:
            # convert numpy array to jpg
            frame_bytes = cv2.imencode('.jpg', self._packet.frame)[1].tobytes()
            robothub.DETECTIONS.send_frame_detection(imagedata=frame_bytes, title=title, camera_serial=self.device.mxid)
        except Exception as e:
            warnings.warn(f'Could not upload detection with error: {e}')


class DepthPacket(HubPacket):
    def __init__(self, device: 'Device', packet):
        super().__init__(device, packet)

    def upload_as_event(self, title):
        raise NotImplementedError('Not implemented yet')


class DetectionPacket(HubPacket):
    def __init__(self, device: 'Device', packet):
        super().__init__(device, packet)
        self.detections = packet.detections

    def upload_as_detection(self, title: str):
        try:
            # convert numpy array to jpg
            frame_bytes = cv2.imencode('.jpg', self._packet.frame)[1].tobytes()

            # TODO add metadata
            robothub.DETECTIONS.send_frame_detection(imagedata=frame_bytes, title=title, camera_serial=self.device.mxid)
        except Exception as e:
            warnings.warn(f'Could not upload detection with error: {e}')

    def upload_as_event(self, title):
        raise NotImplementedError('Not implemented yet')


class TrackerPacket(DetectionPacket):
    def __init__(self, device: 'Device', packet):
        super().__init__(device, packet)
        self.tracklets = packet.tracklets

    def upload_as_event(self, title):
        raise NotImplementedError('Not implemented yet')


class IMUPacket:
    def __init__(self, device: 'Device', packet):
        self.device = device
        self._packet = packet
        # TODO add parsed data



class Streamable:
    """
    Auxiliary class for components that can be streamed to the hub.
    """
    def __init__(self) -> None:
        self.stream_enabled = False
        self.stream_name = None
        self.stream_key = None

    def stream_to_hub(self, name: str, unique_key: str = None) -> None:
        self.stream_enabled = True
        self.stream_name = name
        self.stream_key = unique_key



from abc import abstractmethod, ABC
from dataclasses import asdict
from typing import Callable

import depthai_sdk.classes.packets as packets

from robothub_oak.components.camera import Camera
from robothub_oak.components.neural_network import NeuralNetwork
from robothub_oak.components.stereo import Stereo, DepthQuality, DepthRange
from robothub_oak.hub_camera import HubCamera
from robothub_oak.packets import HubPacket, DetectionPacket, TrackerPacket, DepthPacket, IMUPacket

__all__ = [
    'CreateCameraCommand',
    'CreateNeuralNetworkCommand',
    'CreateStereoCommand',
    'StreamCommand',
    'CommandHistory'
]


class Command(ABC):
    """
    The Command interface declares a method for executing a command.
    """

    def __init__(self, device: 'Device'):
        self.device = device
        self.hub_camera = None

    @abstractmethod
    def execute(self) -> None:
        pass

    def set_camera(self, hub_camera: HubCamera) -> None:
        self.hub_camera = hub_camera

    def get_component(self):
        return None


class CreateCameraCommand(Command):
    """
    Creates a new camera component.
    """

    def __init__(self, device: 'Device', camera: Camera) -> None:
        super().__init__(device=device)
        self._camera = camera

    def execute(self) -> None:
        camera_component = self.hub_camera.create_camera(source=self._camera.name,
                                                         resolution=self._camera.resolution,
                                                         fps=self._camera.fps)
        camera_component.config_color_camera(**asdict(self._camera.camera_config))

        self._camera.camera_component = camera_component

    def get_component(self) -> Camera:
        return self._camera


class CreateNeuralNetworkCommand(Command):
    """
    Creates a new neural network component.
    """

    def __init__(self, device: 'Device', neural_network: NeuralNetwork) -> None:
        super().__init__(device=device)
        self._neural_network = neural_network

    def execute(self) -> None:
        nn_component = self.hub_camera.create_nn(model=self._neural_network.name,
                                                 input=self._neural_network.input.camera_component,
                                                 nn_type=self._neural_network.nn_type,
                                                 tracker=self._neural_network.tracker,
                                                 spatial=self._neural_network.spatial,
                                                 decode_fn=self._neural_network.decode_fn)

        for callback in self._neural_network.callbacks:
            self.hub_camera.callback(nn_component, self._callback_wrapper(callback), True)

        self._neural_network.nn_component = nn_component

    def get_component(self) -> NeuralNetwork:
        return self._neural_network

    def _callback_wrapper(self, callback: Callable) -> Callable[[HubPacket], None]:
        """
        Wraps the callback to be called with a HubPacket.
        :param callback: The callback to be wrapped.
        :return: The wrapped callback.
        """

        def __determine_packet_type(packet) -> Callable:
            packet_type = type(packet)
            if packet_type is packets.DetectionPacket:
                return DetectionPacket
            elif packet_type is packets.TrackerPacket:
                return TrackerPacket
            elif packet_type is packets.DepthPacket:
                return DepthPacket
            elif packet_type is packets.IMUPacket:
                return IMUPacket
            else:
                return HubPacket

        def callback_wrapper(packet):
            callback(__determine_packet_type(packet)(device=self.device, packet=packet))

        return callback_wrapper


class CreateStereoCommand(Command):
    """
    Creates a new stereo component.
    """

    def __init__(self, device: 'Device', stereo: Stereo) -> None:
        super().__init__(device=device)
        self._stereo = stereo

    def execute(self) -> None:
        left = self._stereo.left_camera.camera_component if self._stereo.left_camera else None
        right = self._stereo.right_camera.camera_component if self._stereo.right_camera else None
        stereo_component = self.hub_camera.create_stereo(resolution=self._stereo.resolution,
                                                         fps=self._stereo.fps,
                                                         left=left,
                                                         right=right)

        # Configure stereo component
        stereo_quality = self._stereo.quality
        stereo_range = self._stereo.range

        align = self._stereo.align
        median = 5 if stereo_quality is DepthQuality.DEFAULT else None
        lr_check = stereo_quality is not DepthQuality.FAST
        subpixel = stereo_quality is DepthQuality.QUALITY
        extended_disparity = stereo_range is DepthRange.LONG
        if extended_disparity:
            subpixel = False

        stereo_component.config_stereo(align=align,
                                       lr_check=lr_check,
                                       subpixel=subpixel,
                                       median=median,
                                       extended=extended_disparity)

        self._stereo.stereo_component = stereo_component

    def get_component(self) -> Stereo:
        return self._stereo


class StreamCommand(Command):
    """
    Creates a new stream.
    """

    def __init__(self, device: 'Device', command: 'Command') -> None:
        super().__init__(device=device)
        self._command = command

    def execute(self) -> None:
        component = self._command.get_component()

        if isinstance(component, Camera):
            stream_component = component.camera_component
        elif isinstance(component, NeuralNetwork):
            stream_component = component.nn_component
        elif isinstance(component, Stereo):
            stream_component = component.stereo_component
        else:
            raise Exception('Component not supported for streaming, only Camera and NeuralNetwork are supported.')

        self.hub_camera.create_stream(component=stream_component,
                                      unique_key=component.stream_key,
                                      name=component.stream_name)


class CommandHistory:
    """
    The CommandHistory keeps track of the created commands.
    """

    def __init__(self) -> None:
        self._commands = []

    def push(self, command: Command) -> None:
        """
        Adds a command to the history.
        """
        self._commands.append(command)

    def pop(self) -> Command:
        """
        Removes the last command from the history.
        """
        return self._commands.pop()

    def __len__(self) -> int:
        return len(self._commands)

    def __iter__(self):
        return iter(self._commands)



import contextlib
import logging
import logging as log
import os
from collections import defaultdict
from typing import Optional, List

import robothub

from robothub_oak.device import Device
from robothub_oak.hub_camera import HubCamera

__all__ = ['DeviceManager', 'DEVICE_MANAGER']


class DeviceManager:
    """
    A manager class to handle multiple HubCamera instances.
    """
    REPORT_FREQUENCY = 10  # seconds
    POLL_FREQUENCY = 0.0005

    def __init__(self):
        self._devices = []  # Used to store all devices, shouldn't be modified in any way
        self._hub_cameras = []  # Used to store all HubCamera instances that are currently running

        self.connecting_to_device = False  # Used to prevent multiple threads from connecting to the same device
        self.stop_event = robothub.threading.Event()

        self.reporting_thread = robothub.threading.Thread(target=self._report, name='ReportingThread', daemon=False)
        self.polling_thread = robothub.threading.Thread(target=self._poll, name='PollingThread', daemon=False)
        self.connection_thread = robothub.threading.Thread(target=self._connect, name='ConnectionThread', daemon=False)

    def __exit__(self):
        self.stop()

    def start(self) -> None:
        """
        Start the cameras, start reporting and polling threads.
        """

        # Endless loop to prevent app from exiting if no devices are found
        while not self.stop_event.is_set():
            if self._devices:
                break

            self.stop_event.wait(5)

        self.connection_thread.start()
        log.info('Device connection thread: started successfully.')

        self.reporting_thread.start()
        log.info('Reporting thread: started successfully.')

        self.polling_thread.start()
        log.info('Polling thread: started successfully.')

        # Endless loop to prevent app from exiting, keeps the main thread alive
        while not self.stop_event.is_set():
            self.stop_event.wait(60)

    def stop(self) -> None:
        """
        Stop the cameras, stop reporting and polling threads.
        """
        log.debug('Threads: gracefully stopping...')
        self.stop_event.set()

        self.__graceful_thread_join(self.connection_thread)
        self.__graceful_thread_join(self.reporting_thread)
        self.__graceful_thread_join(self.polling_thread)

        try:
            robothub.STREAMS.destroy_all_streams()
        except BaseException as e:
            logging.debug(f'Destroy all streams excepted with: {e}.')

        for camera in self._hub_cameras:
            try:
                if camera.state != robothub.DeviceState.DISCONNECTED:
                    with open(os.devnull, 'w') as devnull:
                        with contextlib.redirect_stdout(devnull):
                            camera.stop()
            except BaseException as e:
                logging.debug(f'Device {camera.device_name}: could not exit with exception: {e}.')

        log.info('App: stopped successfully.')

    def add_device(self, device: 'Device') -> None:
        """
        Add a camera to the list of cameras.
        """
        self._devices.append(device)

    def remove_device(self, device: 'Device') -> None:
        """
        Remove a camera from the list of cameras.
        """
        self._devices.remove(device)

    @property
    def devices(self) -> List['Device']:
        """
        Returns the list of cameras.
        """
        return self._devices

    def _report(self) -> None:
        """
        Reports the state of the cameras to the agent. Active when app is running, inactive when app is stopped.
        Reporting frequency is defined by REPORT_FREQUENCY.
        """
        while not self.stop_event.is_set():
            for camera in self._hub_cameras:
                try:
                    device_info = camera.info_report()
                    device_stats = camera.stats_report()

                    robothub.AGENT.publish_device_info(device_info)
                    robothub.AGENT.publish_device_stats(device_stats)
                except Exception as e:
                    log.debug(f'Device {camera.device_name}: could not report info/stats with error: {e}.')

            self.stop_event.wait(self.REPORT_FREQUENCY)

    def _poll(self) -> None:
        """
        Polls the cameras for new detections. Polling frequency is defined by POLL_FREQUENCY.
        """
        while not self.stop_event.is_set():
            for camera in self._hub_cameras:
                if not camera.poll():
                    self._disconnect_camera(camera)
                    continue

            self.stop_event.wait(self.POLL_FREQUENCY)

    def _connect(self) -> None:
        """
        Reconnects the cameras that were disconnected or reconnected.
        """
        while not self.stop_event.is_set():
            if len(self._hub_cameras) == len(self._devices) or self.connecting_to_device:
                self.stop_event.wait(5)
                continue

            mxids = [camera.device_name for camera in self._hub_cameras]
            for device in self._devices:
                if device.mxid not in mxids:
                    self.connecting_to_device = True
                    self._connect_device(device)
                    self.connecting_to_device = False

    def _connect_device(self, device: 'Device') -> None:
        """
        Connect a device to the app.
        """
        hub_camera = HubCamera(device_name=device.get_device_name())
        if not device._start(hub_camera):  # Initialize the device (create streams, etc.)
            hub_camera.stop()
            return

        hub_camera.start()  # Start the pipeline

        device.connect_callback(hub_camera)
        self._hub_cameras.append(hub_camera)

        logging.info(f'Device {device.get_device_name()}: started successfully.')

    def _disconnect_camera(self, camera: HubCamera) -> None:
        """
        Disconnect a device from the app.
        """
        log.info(f'Device {camera.device_name}: disconnected.')
        camera.stop()
        self._hub_cameras.remove(camera)

        device = self._get_device_by_name(camera.device_name)
        if device:
            device.disconnect_callback(camera)

    def _get_device_by_name(self, name: str) -> Optional['Device']:
        """
        Get a device by its mxid.
        """
        for device in self._devices:
            if device.get_device_name() == name:
                return device

        return None

    @staticmethod
    def __graceful_thread_join(thread) -> None:
        """
        Gracefully stop a thread.
        """
        try:
            if thread.is_alive():
                thread.join()
        except BaseException as e:
            log.error(f'{thread.getName()}: join excepted with: {e}.')

    @staticmethod
    def get_device(id: str = None, name: str = None, mxid: str = None, ip_address: str = None) -> Device:
        """
        Returns a device by its ID, name, mxid or IP address.

        :param id: The ID of the device.
        :param name: The name of the device.
        :param mxid: The mxid of the device.
        :param ip_address: The IP address of the device.
        :return: The device.
        """
        assert id or name or mxid or ip_address, 'Must specify at least one of id, name, mxid or ip_address'

        device = Device(id=id, name=name, mxid=mxid, ip_address=ip_address)

        # Check if device already exists
        for d in DEVICE_MANAGER.devices:
            if d == device:
                return d

        # Add device to device manager if it doesn't exist
        DEVICE_MANAGER.add_device(device)
        return device

    @staticmethod
    def get_all_devices() -> List[Device]:
        """
        Returns all devices.

        :return: All devices.
        """
        for obj in robothub.DEVICES:
            device_dict = defaultdict(lambda: None, obj.oak)
            device = DEVICE_MANAGER.get_device(
                mxid=device_dict['serialNumber'],
                name=device_dict['productName'],
                id=device_dict['name'],
                ip_address=device_dict['ipAddress']
            )

            exists = False
            for d in DEVICE_MANAGER.devices:
                if d == device:
                    exists = True
                    continue

            if not exists:
                DEVICE_MANAGER.add_device(device)

        return DEVICE_MANAGER.devices


DEVICE_MANAGER = DeviceManager()  # Global device manager



from dataclasses import dataclass, replace
from typing import Optional, Tuple, Union, Any, Dict

import depthai as dai

from robothub_oak.components.streamable import Streamable

__all__ = ['Camera']


@dataclass
class CameraConfig:
    """
    Dataclass representing the configuration of the camera.
    """
    interleaved: Optional[bool] = None
    color_order: Union[None, dai.ColorCameraProperties.ColorOrder, str] = None
    manual_focus: Optional[int] = None
    af_mode: Optional[dai.CameraControl.AutoFocusMode] = None
    awb_mode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None
    scene_mode: Optional[dai.CameraControl.SceneMode] = None
    anti_banding_mode: Optional[dai.CameraControl.AntiBandingMode] = None
    effect_mode: Optional[dai.CameraControl.EffectMode] = None
    isp_scale: Optional[Tuple[int, int]] = None
    sharpness: Optional[int] = None
    luma_denoise: Optional[int] = None
    chroma_denoise: Optional[int] = None


class Camera(Streamable):
    """
    This component represents a single camera on the OAK, either color or mono one.
    The API provides a way to configure the camera, but it is not required to do so.
    """
    def __init__(self, name: str, resolution: Optional[str], fps: Optional[int]) -> None:
        super().__init__()
        self.name = name
        self.resolution = resolution
        self.fps = fps

        self.camera_component = None  # type: depthai_sdk.components.CameraComponent
        self.camera_config = CameraConfig()

    def configure(self,
                  interleaved: Optional[bool] = None,
                  color_order: Union[None, dai.ColorCameraProperties.ColorOrder, str] = None,
                  # Cam control
                  manual_focus: Optional[int] = None,
                  af_mode: Optional[dai.CameraControl.AutoFocusMode] = None,
                  awb_mode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None,
                  scene_mode: Optional[dai.CameraControl.SceneMode] = None,
                  anti_banding_mode: Optional[dai.CameraControl.AntiBandingMode] = None,
                  effect_mode: Optional[dai.CameraControl.EffectMode] = None,
                  # IQ settings
                  isp_scale: Optional[Tuple[int, int]] = None,
                  sharpness: Optional[int] = None,
                  luma_denoise: Optional[int] = None,
                  chroma_denoise: Optional[int] = None,
                  ) -> None:
        """
        Configures the camera component.
        """
        kwargs = self._process_kwargs(locals())

        if len(kwargs) > 0:
            self.camera_config = replace(self.camera_config, **kwargs)

    def set_resolution(self, resolution: str) -> None:
        """
        Sets the resolution of the camera.

        :param resolution: String representation of the resolution, e.g. '1080p' or '4K'.
        """
        self.resolution = resolution

    def set_fps(self, fps: int) -> None:
        """
        Sets the FPS of the camera.

        :param fps: FPS to set as an integer.
        """
        self.fps = fps

    @staticmethod
    def _process_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Process the kwargs and remove all None values."""
        kwargs.pop('self')
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return kwargs



from typing import Union, Callable, Optional

from robothub_oak.components import Camera
from robothub_oak.components.streamable import Streamable

__all__ = ['NeuralNetwork']


class NeuralNetwork(Streamable):
    def __init__(self,
                 name: str,
                 input: Union[Camera, 'NeuralNetwork'],
                 nn_type: Optional[str] = None,  # Either 'yolo' or 'mobilenet'
                 decode_fn: Optional[Callable] = None,
                 tracker: bool = False,
                 spatial: Optional[bool] = None):
        super().__init__()
        self.name = name
        self.input = input
        self.nn_type = nn_type
        self.decode_fn = decode_fn
        self.tracker = tracker
        self.spatial = spatial

        self.nn_component = None  # type: depthai_sdk.components.NNComponent
        self.callbacks = []

    def add_callback(self, callback: Callable) -> None:
        self.callbacks.append(callback)

    def configure(self, **kwargs) -> None:
        pass



from enum import IntEnum
from typing import Union, Optional

from robothub_oak.components.streamable import Streamable

__all__ = ['Stereo', 'DepthQuality', 'DepthRange']


class DepthQuality(IntEnum):
    FAST = 1  # nothing turned on
    DEFAULT = 2  # lr check, median filter
    QUALITY = 3  # lr check, subpixel


class DepthRange(IntEnum):
    SHORT = 1  # extended disparity
    DEFAULT = 2  # no effect
    LONG = 3  # subpixel


class Stereo(Streamable):
    def __init__(self,
                 resolution: Optional[str],
                 fps: Optional[int],
                 left_camera: Optional['Camera'] = None,
                 right_camera: Optional['Camera'] = None):
        """
        This component represents the stereo module of the OAK device.

        :param resolution: Resolution of the disparity/depth map. Can be one of '400p', '480p', '720p', or '800p'.
        :param fps: FPS of the stereo output.
        :param left_camera: Left camera component.
        :param right_camera: Right camera component.
        """
        super().__init__()
        self.resolution = resolution
        self.fps = fps

        self.left_camera = left_camera
        self.right_camera = right_camera

        self.quality = DepthQuality.DEFAULT
        self.range = DepthRange.DEFAULT

        self.align = 'color'
        self.stereo_component = None  # type: depthai_sdk.components.StereoComponent

    def configure(self,
                  depth_quality: Union[str, DepthQuality] = None,
                  depth_range: Union[str, DepthRange] = None,
                  align: str = None) -> None:
        """
        Configures the stereo component.

        :param depth_quality: Quality of the depth map. Can be one of 'fast', 'default' or 'quality'.
        :param depth_range: Working range of the stereo module. Can be one of 'short', 'default' or 'long'.
        :param align: Alignment of the depth map. Can be one of 'color', 'left', 'right' or 'cama,c' (or similar). Defaults to 'color'.
        """
        if depth_quality:
            self.quality = self._set_enum_value(DepthQuality, depth_quality)
        if depth_range:
            self.range = self._set_enum_value(DepthRange, depth_range)
        if align:
            self.align = align

    @staticmethod
    def _set_enum_value(enum, value):
        if isinstance(value, str):
            return enum[value.upper()]
        elif isinstance(value, enum):
            return value
        else:
            raise ValueError(f'Invalid value: {value}')



import logging

__all__ = ['try_or_default', 'set_logging_level']


def try_or_default(func, default=None):
    try:
        return func()
    except:
        return default


def set_logging_level(level):
    logging.basicConfig()
    logging.getLogger().setLevel(level)



import warnings
from typing import Callable, Any, Optional, Dict

from depthai import NNData

from robothub_oak.commands import (
    CreateStereoCommand, CreateCameraCommand, CreateNeuralNetworkCommand,
    StreamCommand, CommandHistory
)
from robothub_oak.components.camera import Camera
from robothub_oak.components.neural_network import NeuralNetwork
from robothub_oak.components.stereo import Stereo
from robothub_oak.components.streamable import Streamable
from robothub_oak.hub_camera import HubCamera

__all__ = ['Device']


class Device:
    """
    Device represents a single device. It is used to create components and execute commands.
    Startup and shutdown of the device is handled by the DeviceManager.
    """

    def __init__(self, id: str = None, name: str = None, mxid: str = None, ip_address: str = None) -> None:
        """
        :param id: ID of the device.
        :param name: Name of the device.
        :param mxid: MXID of the device.
        :param ip_address: IP address of the device.
        """
        # Device info
        self.id = id  # device identifier provided by the user
        self.name = name  # product name
        self.mxid = mxid  # mxid of the device
        self.ip_address = ip_address  # IP address of the device

        self.cameras: Dict[str, Camera] = {}
        self.stereo: Optional[Stereo] = None
        self.neural_networks: Dict[str, NeuralNetwork] = {}

        self.hub_camera: Optional[HubCamera] = None

        # Callbacks
        self.disconnect_callback: Callable[[Any], None] = lambda x: None
        self.connect_callback: Callable[[Any], None] = lambda x: None

        self._command_history = CommandHistory()

    def __eq__(self, other):
        if isinstance(other, Device):
            return self.id == other.id or self.name == other.name or self.mxid == other.mxid or self.ip_address == other.ip_address
        elif isinstance(other, str):
            return self.id == other or self.name == other or self.mxid == other or self.ip_address == other
        else:
            return False

    def _start(self, hub_camera: HubCamera) -> bool:
        """
        Internal method to execute all commands.

        :param hub_camera: The HubCamera instance to use.
        :return: True if successful, False otherwise.
        """
        if hub_camera.oak_camera is None:
            return False

        self.hub_camera = hub_camera

        try:
            for command in self._command_history:
                if isinstance(command, CreateStereoCommand) and not hub_camera.has_stereo:
                    warnings.warn(f'Device {self.get_device_name()} does not support stereo, skipping stereo creation.')
                    continue

                command.set_camera(hub_camera)
                command.execute()
        except Exception as e:
            warnings.warn(f'Failed to start device {self.get_device_name()} with error: {e}')
            return False

        # Create streams
        for command in self._command_history:
            # Check if stream is enabled, if so, create a stream command and execute it
            component = command.get_component()

            # Component can be None if the command failed to execute (e.g., stereo component on a single camera device)
            if isinstance(command, CreateStereoCommand) and component.stereo_component is None:
                continue

            if isinstance(component, Streamable) and component.stream_enabled:
                stream_command = StreamCommand(self, command)
                stream_command.set_camera(hub_camera)
                stream_command.execute()

        return True

    def restart(self) -> bool:
        """
        Restarts the device. This will stop all components and streams and recreate them.

        :return: True if successful, False otherwise.
        """
        try:
            if self.hub_camera:
                self.hub_camera.stop()
                self.hub_camera.oak_camera = self.hub_camera.init_oak_camera()
                self._start(hub_camera=self.hub_camera)
        except Exception as e:
            warnings.warn(f'Failed to restart device {self.get_device_name()} with error: {e}')
            return False

        return True

    def get_camera(self, name: str, resolution: str = None, fps: int = None) -> Camera:
        """
        Creates a camera component.

        :param name: The name of the camera.
        :param resolution: The resolution of the camera.
        :param fps: The FPS of the camera.
        :return: The camera.
        """
        # Check if camera already exists
        if name in self.cameras:
            return self.cameras[name]

        camera = Camera(name, resolution, fps)
        self.cameras[name] = camera

        command = CreateCameraCommand(self, camera)
        self._command_history.push(command)
        return camera

    def create_neural_network(self,
                              name: str,
                              input: Camera = None,
                              nn_type: str = None,
                              decode_fn: Callable[[NNData], Any] = None,
                              tracker: bool = False,
                              spatial: Optional[bool] = None
                              ) -> NeuralNetwork:
        """
        Creates a neural network.

        :param name: The name of the neural network.
        :param input: The input camera.
        :param nn_type: The type of neural network. Either 'yolo' or 'mobilenet'.
        :param decode_fn: The decode function to use. Decoding is done on the host.
        :param tracker: Whether to use tracking.
        :param spatial: Whether to use spatial detection.
        :return: The neural network.
        """
        if name in self.neural_networks:
            return self.neural_networks[name]
        elif not input:
            raise ValueError('Neural network must have an input')

        if isinstance(input, NeuralNetwork):
            raise NotImplementedError('Neural networks cannot be used as input for other neural networks yet')

        neural_network = NeuralNetwork(name=name,
                                       input=input,
                                       nn_type=nn_type,
                                       decode_fn=decode_fn,
                                       tracker=tracker,
                                       spatial=spatial)
        command = CreateNeuralNetworkCommand(self, neural_network)
        self._command_history.push(command)
        return neural_network

    def get_stereo_camera(self,
                          resolution: str = None,
                          fps: int = None,
                          left_camera: Camera = None,
                          right_camera: Camera = None
                          ) -> Stereo:
        """
        Creates a stereo component.

        :param resolution: The resolution of the stereo camera.
        :param fps: The FPS of the stereo camera.
        :param left_camera: The left camera.
        :param right_camera: The right camera.
        """
        if self.stereo:
            return self.stereo

        self.stereo = Stereo(resolution, fps, left_camera, right_camera)
        command = CreateStereoCommand(self, self.stereo)
        self._command_history.push(command)
        return self.stereo

    def set_connect_callback(self, callback: Callable[[HubCamera], None]) -> None:
        """
        Sets the callback to be called when the device connects.

        :param callback: The callback to be called when the device connects.
        """
        self.connect_callback = callback

    def set_disconnect_callback(self, callback: Callable[[HubCamera], None]) -> None:
        """
        Sets the callback to be called when the device disconnects.

        :param callback: The callback to be called when the device disconnects.
        """
        self.disconnect_callback = callback

    def get_device_name(self) -> str:
        """
        Returns the name of the device.
        """
        return self.id or self.name or self.ip_address or self.mxid



import json
import time
from functools import partial
from typing import Callable

from robothub import StreamHandle

__all__ = [
    'get_default_color_callback',
    'get_default_nn_callback',
    'get_default_depth_callback',
]


def get_default_color_callback(stream_handle: StreamHandle) -> Callable:
    """
    Returns a default callback for color streams.

    :param stream_handle: StreamHandle instance to publish the data to.
    """
    return partial(_default_encoded_callback, stream_handle)


def get_default_nn_callback(stream_handle: StreamHandle) -> Callable:
    """
    Returns a default callback for NN streams.

    :param stream_handle: StreamHandle instance to publish the data to.
    """
    return partial(_default_nn_callback, stream_handle)


def get_default_depth_callback(stream_handle: StreamHandle) -> Callable:
    """
    Returns a default callback for depth streams.

    :param stream_handle: StreamHandle instance to publish the data to.
    """
    return partial(_default_encoded_callback, stream_handle)


def _default_encoded_callback(stream_handle: StreamHandle, packet):
    """
    Default callback for encoded streams.

    :param stream_handle: StreamHandle instance to publish the data to.
    :param packet: Packet instance containing the data.
    """

    timestamp = int(time.time() * 1_000)
    frame_bytes = bytes(packet.msg.getData())
    stream_handle.publish_video_data(frame_bytes, timestamp, None)


def _default_nn_callback(stream_handle: StreamHandle, packet):
    """
    Default callback for NN streams.

    :param stream_handle: StreamHandle instance to publish the data to.
    :param packet: Packet instance containing the data.
    """
    visualizer = packet.visualizer
    metadata = None
    if visualizer:
        metadata = json.loads(visualizer.serialize())
        visualizer.reset()

        # temp fix to replace None value that causes errors on frontend
        if not metadata['config']['detection']['color']:
            metadata['config']['detection']['color'] = [255, 0, 0]

    timestamp = int(time.time() * 1_000)
    frame_bytes = bytes(packet.msg.getData())
    stream_handle.publish_video_data(frame_bytes, timestamp, metadata)



import logging as log
import time
import warnings
from pathlib import Path
from typing import Union, Optional, Callable, Dict, Any

import depthai
import depthai as dai
import depthai_sdk
import robothub
from depthai_sdk import OakCamera
from depthai_sdk.components import CameraComponent, StereoComponent, NNComponent

from robothub_oak.callbacks import get_default_color_callback, get_default_nn_callback, get_default_depth_callback
from robothub_oak.utils import try_or_default

__all__ = ['HubCamera']


class HubCamera:
    """
    Wrapper for the DepthAI OakCamera class.
    """

    def __init__(self,
                 device_name: str,
                 usb_speed: Union[None, str, dai.UsbSpeed] = None,
                 rotation: int = 0):
        """
        :param device_name: Device identifier, either mxid, IP address or USB port.
        :param usb_speed: USB speed to use.
        :param rotation: Rotation of the camera, defaults to 0.
        """
        self.state = robothub.DeviceState.UNKNOWN
        self.device_name = device_name
        self.usb_speed = usb_speed
        self.rotation = rotation

        self.stop_event = robothub.threading.Event()
        self.streams = {}  # unique_key -> StreamHandle
        self.available_sensors = []  # self.oak_camera.sensors if self.oak_camera else []

        self.oak_camera = self.init_oak_camera()

    def init_oak_camera(self) -> Optional[OakCamera]:
        """
        Initializes the OakCamera instance. Will try to initialize for 5 seconds before returning None.

        :return: OakCamera instance if successful, None otherwise.
        """
        start_time = time.time()
        while not self.stop_event.is_set():
            try:
                camera = OakCamera(self.device_name, usb_speed=self.usb_speed, rotation=self.rotation)
                self.available_sensors = camera.sensors
                return camera
            except Exception:
                if time.time() - start_time > 10:
                    break

                self.stop_event.wait(1)

        return None

    def create_camera(self,
                      source: str,
                      resolution: Union[
                          None, str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
                      ] = None,
                      fps: Optional[float] = None
                      ) -> CameraComponent:
        """
        Creates a camera component.

        :param source: Source of the camera. Can be either 'color', 'left', 'right' or a sensor name.
        :param resolution: Resolution of the camera.
        :param fps: FPS of the output stream.
        :return: The camera component.
        """
        comp = self.oak_camera.create_camera(source=source, resolution=resolution, fps=fps, encode='h264')
        return comp

    def create_nn(self,
                  model: Union[str, Path],
                  input: Union[CameraComponent, NNComponent],
                  nn_type: Optional[str] = None,
                  tracker: bool = False,
                  spatial: Union[None, bool, StereoComponent] = None,
                  decode_fn: Optional[Callable] = None
                  ) -> NNComponent:
        """
        Creates a neural network component.

        :param model: Name or path to the model.
        :param input: Input component, either a camera or another neural network.
        :param nn_type: Either 'yolo' or 'mobilenet'. For other types, use None.
        :param tracker: If True, will enable and add a tracker to the output.
        :param spatial: If True, will enable and add spatial data to the output. If a StereoComponent is provided, will use that component to calculate the spatial data.
        :param decode_fn: Function to decode the output of the neural network.

        :return: Neural network component.
        """
        comp = self.oak_camera.create_nn(model=model, input=input, nn_type=nn_type,
                                         tracker=tracker, spatial=spatial, decode_fn=decode_fn)
        return comp

    def create_stereo(self,
                      resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
                      fps: Optional[float] = None,
                      left: Union[None, dai.Node.Output, depthai_sdk.components.CameraComponent] = None,
                      right: Union[None, dai.Node.Output, depthai_sdk.components.CameraComponent] = None,
                      ) -> StereoComponent:
        """
        Creates a stereo component.

        :param resolution: Resolution of the stereo component.
        :param fps: FPS of the output stream.
        :param left: Left camera component, optional.
        :param right: Right camera component, optional.
        :return: Stereo component.
        """
        if not self.has_stereo:
            raise RuntimeError('Device does not have stereo cameras.')

        comp = self.oak_camera.create_stereo(resolution=resolution, fps=fps, left=left, right=right, encode='h264')
        comp.set_colormap(dai.Colormap.STEREO_TURBO)
        return comp

    def create_stream(self,
                      component: Union[CameraComponent, NNComponent, StereoComponent],
                      unique_key: str,
                      name: str,
                      callback: Callable = None
                      ) -> None:
        """
        Creates a stream for the given component.

        :param component: Component to create a stream for.
        :param unique_key: Unique key for the stream.
        :param name: Name of the stream that will be used in Live View.
        :param callback: Callback function to be called when a new frame is received.
        """
        log.debug(f'Stream: creating stream {name} for component {component}.')

        if unique_key is None:
            unique_key = f'{self.device_name}_{component.__class__.__name__.lower()}_{component.out.encoded.__name__}'

        if unique_key in robothub.STREAMS.streams.keys():
            stream_handle = robothub.STREAMS.streams[unique_key]
        else:
            stream_handle = robothub.STREAMS.create_video(camera_serial=self.device_name,
                                                          unique_key=unique_key,
                                                          description=name)
        self.streams[unique_key] = stream_handle

        self._add_stream_callback(stream_handle=stream_handle, component=component, callback=callback)

    def _add_stream_callback(self,
                             stream_handle: robothub.StreamHandle,
                             component: Union[CameraComponent, NNComponent, StereoComponent],
                             callback: Callable
                             ) -> None:
        """
        Selects the correct callback function for the given component and adds it to the stream.

        :param stream_handle: Stream handle to add the callback to.
        :param component: Component to create a callback for.
        :param callback: User-defined callback function to be called when a new frame is received.
        """
        fn = None
        enable_visualizer = False
        if isinstance(component, CameraComponent):
            fn = callback or get_default_color_callback(stream_handle)
        elif isinstance(component, NNComponent):
            fn = callback or get_default_nn_callback(stream_handle)
            enable_visualizer = True
        elif isinstance(component, StereoComponent):
            fn = callback or get_default_depth_callback(stream_handle)

        if fn:
            self.oak_camera.callback(component.out.encoded, callback=fn, enable_visualizer=enable_visualizer)

    def callback(self, output: Any, callback: Callable, enable_visualizer: bool = False) -> None:
        """
        Sets a callback function for the given output.

        :param output: Output to set the callback for.
        :param callback: Callback function to be called when a new frame is received.
        :param enable_visualizer: Whether to enable the visualizer that provides metadata.
        """
        self.oak_camera.callback(output, callback=callback, enable_visualizer=enable_visualizer)

    def poll(self) -> Optional[int]:
        """
        Polls the device for new data.
        """
        return self.oak_camera.poll()

    def start(self) -> None:
        """
        Starts the device and sets the state to connected.
        """
        if self.state == robothub.DeviceState.CONNECTED:
            return

        while not self.stop_event.is_set():
            try:
                self.oak_camera.start()
                self.state = robothub.DeviceState.CONNECTED
                return
            except Exception as e:
                warnings.warn(f'Camera: could not start with exception {e}.')

            self.stop_event.wait(1)

    def stop(self) -> None:
        """
        Stops the device and sets the state to disconnected.
        """
        self.stop_event.set()

        for stream in self.streams.values():
            try:
                robothub.STREAMS.destroy(stream)
            except ValueError:
                pass

        self.streams.clear()
        self.state = robothub.DeviceState.DISCONNECTED

        self.streams.clear()

        if self.oak_camera:
            self.oak_camera.device.close()

        self.oak_camera = None
        self.available_sensors = []

    def stats_report(self) -> Dict[str, Any]:
        """
        Returns a dictionary with statistics about the device.
        """
        stats = {'mxid': self.device.getMxId()}

        css_cpu_usage = self.device.getLeonCssCpuUsage().average
        mss_cpu_usage = self.device.getLeonMssCpuUsage().average
        cmx_mem_usage = self.device.getCmxMemoryUsage()
        ddr_mem_usage = self.device.getDdrMemoryUsage()
        chip_temp = self.device.getChipTemperature()

        stats['css_usage'] = int(100 * css_cpu_usage)
        stats['mss_usage'] = int(100 * mss_cpu_usage)
        stats['ddr_mem_free'] = int(ddr_mem_usage.total - ddr_mem_usage.used)
        stats['ddr_mem_total'] = int(ddr_mem_usage.total)
        stats['cmx_mem_free'] = int(cmx_mem_usage.total - cmx_mem_usage.used)
        stats['cmx_mem_total'] = int(cmx_mem_usage.total)
        stats['css_temp'] = int(100 * chip_temp.css)
        stats['mss_temp'] = int(100 * chip_temp.mss)
        stats['upa_temp'] = int(100 * chip_temp.upa)
        stats['dss_temp'] = int(100 * chip_temp.dss)
        stats['temp'] = int(100 * chip_temp.average)

        return stats

    def info_report(self) -> Dict[str, Any]:
        """
        Returns a dictionary with information about the device.
        """
        info = {
            'mxid': self.device.getMxId(),
            'protocol': 'unknown',
            'platform': 'unknown',
            'product_name': 'unknown',
            'board_name': 'unknown',
            'board_rev': 'unknown',
            'bootloader_version': 'unknown',
            'state': self.state.value,
        }

        device_info = try_or_default(self.device.getDeviceInfo)
        calibration = try_or_default(self.device.readFactoryCalibration) or try_or_default(self.device.readCalibration2)
        eeprom_data = try_or_default(calibration.getEepromData)

        bootloader_version = self.device.getBootloaderVersion()  # can be None
        if bootloader_version:
            info['bootloader_version'] = bootloader_version.toStringSemver()

        if eeprom_data:
            info['product_name'] = eeprom_data.productName
            info['board_name'] = eeprom_data.boardName
            info['board_rev'] = eeprom_data.boardRev

        if device_info:
            info['protocol'] = device_info.protocol.name
            info['platform'] = device_info.platform.name

        return info

    @property
    def device(self) -> dai.Device:
        """
        Returns the device object.
        :return: depthai.Device object.
        """
        return self.oak_camera.device

    @property
    def is_connected(self) -> bool:
        """
        Returns whether the device is connected or not.
        :return: True if connected, False otherwise.
        """
        return not self.device.isClosed()

    @property
    def has_color(self) -> bool:
        """
        Returns whether the device has a color camera.
        :return: True if the device has a color camera, False otherwise.
        """
        return depthai.CameraBoardSocket.RGB in self.available_sensors

    @property
    def has_left(self) -> bool:
        """
        Returns whether the device has a left camera.
        :return: True if the device has a left camera, False otherwise.
        """
        return depthai.CameraBoardSocket.LEFT in self.available_sensors

    @property
    def has_right(self) -> bool:
        """
        Returns whether the device has a right camera.
        :return: True if the device has a right camera, False otherwise.
        """
        return depthai.CameraBoardSocket.RIGHT in self.available_sensors

    @property
    def has_stereo(self) -> bool:
        """
        Returns whether the device has a depth camera.
        :return: True if the device has a stereo camera, False otherwise.
        """
        return self.has_left and self.has_right
